/*
 * tty.c
 * terminal handling routines for powwow
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <unistd.h>

#ifdef USE_TERMIO
#  ifdef APOLLO
#    include "/sys5.3/usr/include/sys/termio.h"
#  else
/*
 * including both termio.h and termios.h might be an overkill, and gives
 * many warnings, but seems to be necessary at times. works anyway.
 */
#    include <termios.h>
#    include <termio.h>
#  endif
/* #else not USE_TERMIO */
#endif

/*
 * SunOS 4 doesn't have function headers and has the defs needed from
 * ioctl.h in termios.h.  Does it compile without USE_TERMIO?
 */
#if (defined(sun) && defined(sparc) && ! defined(__SVR4))
  extern int printf();
#else
  #include <sys/ioctl.h>
#endif

#ifdef BSD_386
#  include <sys/ioctl_compat.h>
#  define O_RAW RAW
#  define O_ECHO ECHO
#  define O_CBREAK CBREAK
#endif

#if defined(TCSETS) || defined(TCSETATTR)
#  ifndef TCSETS		/* cc for HP-UX  SHOULD define this... */
#    define TCSETS TCSETATTR
#    define TCGETS TCGETATTR
#  endif
typedef struct termios termiostruct;
#else
#  define TCSETS TCSETA
#  define TCGETS TCGETA
typedef struct termio termiostruct;
#endif

#ifdef VSUSP
#  define O_SUSP VSUSP
#else
#  ifdef SWTCH
#    define O_SUSP SWTCH
#  else
#    define O_SUSP SUSP
#  endif
#endif

#include "powwow.h"
#include "edit.h"
#include "list.h"

/* int ioctl(); */

#ifdef USE_VT100	/* hard-coded vt100 features if no termcap: */

static char kpadstart[] = "", kpadend[] = "", begoln[] = "\r",
	clreoln[] = "\033[K", clreoscr[] = "\033[J",
	leftcur[] = "\033[D", rightcur[] = "\033[C",  upcur[] = "\033[A",
	modebold[] = "\033[1m", modeblink[] = "\033[5m", modeinv[] = "\033[7m",
	modeuline[] = "\033[4m", modestandon[] = "", modestandoff[] = "",
        modenorm[] = "\033[m", modenormbackup[4],
	cursor_left[] = "\033[D", cursor_right[] = "\033[C", 
        cursor_up[] = "\033[A", cursor_down[] = "\033[B";
	
#define insertfinish (0)
static int len_begoln = 1, len_leftcur = 3, len_upcur = 3, gotocost = 8;

#else /* not USE_VT100, termcap function declarations */

int tgetent();
int tgetnum();
int tgetflag();
char *tgetstr();
char *tgoto();

/* terminal escape sequences */
static char kpadstart[CAPLEN], kpadend[CAPLEN],
	leftcur[CAPLEN], rightcur[CAPLEN], upcur[CAPLEN], curgoto[CAPLEN],
	delchar[CAPLEN], insstart[CAPLEN], insstop[CAPLEN], 
        inschar[CAPLEN],
	begoln[CAPLEN], clreoln[CAPLEN], clreoscr[CAPLEN],
	cursor_left[CAPLEN], cursor_right[CAPLEN], cursor_up[CAPLEN],
	cursor_down[CAPLEN];

/* attribute changers: */
static char modebold[CAPLEN], modeblink[CAPLEN], modeinv[CAPLEN],
            modeuline[CAPLEN], modestandon[CAPLEN], modestandoff[CAPLEN],
            modenorm[CAPLEN], modenormbackup[CAPLEN];

static int len_begoln, len_clreoln, len_leftcur, len_upcur, gotocost,
           deletecost, insertcost, insertfinish, inscharcost;

static int extract();

#endif /* USE_VT100 */


char *tty_modebold = modebold,       *tty_modeblink = modeblink,
     *tty_modeinv  = modeinv,        *tty_modeuline = modeuline,
     *tty_modestandon = modestandon, *tty_modestandoff = modestandoff,
     *tty_modenorm = modenorm,       *tty_modenormbackup = modenormbackup,
     *tty_begoln   = begoln,         *tty_clreoln = clreoln,
     *tty_clreoscr = clreoscr;

int tty_read_fd = 0, tty_output_fd = 1;


#ifdef USE_TERMIO
static termiostruct ttybsave;
#else /* not USE_TERMIO */
static struct sgttyb ttybsave;
static struct tchars tcsave;
static struct ltchars ltcsave;
#endif /* USE_TERMIO */

/*
 * Terminal handling routines:
 * These are one big mess of left-justified chicken scratches.
 * It should be handled more cleanly...but unix portability is what it is.
 */

/*
 * puts() wrapper, does not append '\n' to string
 */
static void puts_wrapper(s)
char *s;
{
    printf("%s", s); /* fputs(s, stdout); would be equally good */
}
void (*tty_puts)() = puts_wrapper;

/*
 * putc() wrapper
 */
static void putc_wrapper(c)
char c;
{
    (void)putc((int)(unsigned char)c, stdout);
}
void (*tty_putc)() = putc_wrapper;

static void flush_wrapper()
{
    fflush(stdout);
}
void (*tty_flush)() = flush_wrapper;
void (*tty_printf)() = (void (*)())printf;
int (*tty_read)() = (int (*)())read;

static void gets_wrapper(s, size)
char *s;
int size;
{
    fgets(s, size, stdin);
}
void (*tty_gets)() = gets_wrapper;

/*
 * Set the terminal to character-at-a-time-without-echo mode, and save the
 * original state in ttybsave
 */
static void init()
{
#ifdef USE_TERMIO
    termiostruct ttyb;
    ioctl(tty_read_fd, TCGETS, &ttyb);
    ttybsave = ttyb;
    ttyb.c_lflag &= ~(ECHO|ICANON);
    ttyb.c_cc[VTIME] = 0;
    ttyb.c_cc[VMIN] = 1;
    /* disable the special handling of the suspend key (handle it ourselves) */
    ttyb.c_cc[O_SUSP] = 0;
    ioctl(tty_read_fd, TCSETS, &ttyb);
#else /* not USE_TERMIO */
    struct sgttyb ttyb;
    struct ltchars ltc;
    ioctl(tty_read_fd, TIOCGETP, &ttybsave);
    ioctl(tty_read_fd, TIOCGETC, &tcsave);
    ioctl(tty_read_fd, TIOCGLTC, &ltcsave);
    ttyb = ttybsave;
    ttyb.sg_flags = (ttyb.sg_flags|O_CBREAK) & ~O_ECHO;
    ioctl(tty_read_fd, TIOCSETP, &ttyb);
    ltc = ltcsave;
    ltc.t_suspc = -1;
    ioctl(tty_read_fd, TIOCSLTC, &ltc);
#endif /* USE_TERMIO */

    tty_puts(kpadstart);
    tty_flush();
    
    setvbuf(stdout, NULL, _IOFBF, BUFSIZ);
}
void (*tty_init)() = init;

/*
 * Reset the terminal to its original state
 */
static void quit()
{
#ifdef USE_TERMIO
    ioctl(tty_read_fd, TCSETS, &ttybsave);
#else  /* not USE_TERMIO */
    ioctl(tty_read_fd, TIOCSETP, &ttybsave);
    ioctl(tty_read_fd, TIOCSETC, &tcsave);
    ioctl(tty_read_fd, TIOCSLTC, &ltcsave);
#endif /* USE_TERMIO */
    tty_puts(kpadend);
    tty_flush();
}
void (*tty_quit)() = quit;

/*
 * enable/disable special keys depending on the current linemode
 */
static void special_keys()
{
#ifdef USE_TERMIO
    int i;
    termiostruct ttyb;
    ioctl(tty_read_fd, TCGETS, &ttyb);
    if (linemode & LM_CHAR)  {
	/* char-by-char mode: disable all special keys and set raw mode */
	for(i = 0; i < NCCS; i++)
	    ttyb.c_cc[i] = 0;
	ttyb.c_oflag &= ~OPOST;
    } else {
	/* line at a time mode: enable them, except suspend */
	for(i = 0; i < NCCS; i++)
	    ttyb.c_cc[i] = ttybsave.c_cc[i];
	/* disable the suspend key (handle it ourselves) */
	ttyb.c_cc[O_SUSP] = 0;
	/* set cooked mode */
	ttyb.c_oflag |= OPOST;
    }
    ioctl(tty_read_fd, TCSETS, &ttyb);
#else /* not USE_TERMIO */
    struct tchars tc = {-1, -1, -1, -1, -1, -1};
    struct ltchars ltc = {-1, -1, -1, -1, -1, -1};
    struct sgttyb ttyb;
    ioctl(tty_read_fd, TIOCGETP, &ttyb);
    if (linemode & LM_CHAR) {
	/* char-by-char mode: set RAW mode*/
	ttyb.sg_flags |= RAW;
    } else {
	/* line-at-a-time mode: enable spec keys, disable RAW */
	tc = tcsave;
	ltc = ltcsave;
	ltc.t_suspc = -1;	/* suspend key remains disabled */
	ttyb.sg_flags &= ~RAW;
    }
    ioctl(tty_read_fd, TIOCSETP, &ttyb);
    ioctl(tty_read_fd, TIOCSETC, &tc);
    ioctl(tty_read_fd, TIOCSLTC, &ltc);
#endif /* USE_TERMIO */
}
void (*tty_special_keys)() = special_keys;


/*
 * read termcap definitions
 */
static void bootsrap()
{
#ifndef USE_VT100
    struct tc_init_node {
	char cap[4], *buf;
	int *len, critic;
    };
    static struct tc_init_node tc_init[] = {
	{ "cm", curgoto, 0, 1 },
	{ "ce", clreoln, &len_clreoln, 1 },
	{ "cd", clreoscr, 0, 1 },
	{ "nd", rightcur, 0, 1 },
	{ "le", leftcur, &len_leftcur, 0 },
	{ "up", upcur,   &len_upcur,  0 },
	{ "cr", begoln,  &len_begoln, 0 },
	{ "ic", inschar, &inscharcost, 0 },
	{ "im", insstart, &insertcost, 0 },
	{ "ei", insstop, &insertcost, 0 },
	{ "dm", delchar, &deletecost, 0 },
	{ "dc", delchar, &deletecost, 0 },
	{ "ed", delchar, &deletecost, 0 },
	{ "me", modenorm, 0, 0 },
	{ "md", modebold, 0, 0 },
	{ "mb", modeblink, 0, 0 },
	{ "mr", modeinv, 0, 0 },
	{ "us", modeuline, 0, 0 },
	{ "so", modestandon, 0, 0 },
	{ "se", modestandoff, 0, 0 },
	{ "ks", kpadstart, 0, 0 },
	{ "ke", kpadend, 0, 0 },
	{ "kl", cursor_left, 0, 0 },
	{ "kr", cursor_right, 0, 0 },
	{ "ku", cursor_up, 0, 0 },
	{ "kd", cursor_down, 0, 0 },
	{ "", NULL, 0, 0 }
    };
    struct tc_init_node *np;
    char tcbuf[2048];		/* by convention, this is enough */
    int i;
#endif /* not USE_VT100 */
#if !defined(USE_VT100) || defined(BUG_TELNET)
    char *term = getenv("TERM");
    if (!term) {
	fprintf(stderr, "$TERM not set\n");
	exit(1);
    }
#endif /* !defined(USE_VT100) || defined(BUG_TELNET) */
#ifdef USE_VT100
    cols = 80;
#  ifdef LINES
    lines = LINES;
#  else  /* not LINES */
    lines = 24;
#  endif /* LINES */
#else   /* not USE_VT100 */
    switch(tgetent(tcbuf, term)) {
     case 1:
	break;
     case 0:
	fprintf(stderr,
		"There is no entry for `%s' in the terminal data base.\n", term);
	fprintf(stderr,
		"Please set your $TERM environment variable correctly.\n");
	exit(1);
     default:
	syserr("tgetent");
    }
    for(np = tc_init; np->cap[0]; np++)
	if ((i = extract(np->cap, np->buf))) {
	    if (np->len) *np->len += i;
	} else if (np->critic) {
	    fprintf(stderr,
		    "Your terminal is not powerful enough, missing `%s'.\n",
		    np->cap);
	    exit(1);
	}
    if (!len_begoln)
	strcpy(begoln, "\r"), len_begoln = 1;
    if (!len_leftcur)
	strcpy(leftcur, "\b"), len_leftcur = 1;
    cols = tgetnum("co");
    lines = tgetnum("li");
    gotocost = strlen(tgoto(curgoto, cols - 1, lines - 1));
    insertfinish = gotocost + len_clreoln;
#endif  /* not USE_VT100 */
    strcpy(modenormbackup, modenorm);
#ifdef BUG_TELNET
    if (strncmp(term, "vt10", 4) == 0) {
	/* might be NCSA Telnet 2.2 for PC, which doesn't reset colours */
	sprintf(modenorm, "\033[;%c%d;%s%dm", 
		DEFAULTFG<LOWCOLORS ? '3' : '9',  DEFAULTFG % LOWCOLORS,
		DEFAULTBG<LOWCOLORS ? "4" : "10", DEFAULTBG % LOWCOLORS);
    }
#endif /* BUG_TELNET */
}
void (*tty_bootsrap)() = bootsrap;

/*
 * add the default keypad bindings to the list
 */
static void add_walk_binds()
{
    /*
     * Note: termcap doesn't have sequences for the numeric keypad, so we just
     * assume they are the same as for a vt100. They can be redefined
     * at runtime anyway (using #bind or #rebind)
     */
    add_keynode("KP2", "\033Or", 0, key_run_command, "s");
    add_keynode("KP3", "\033Os", 0, key_run_command, "d");
    add_keynode("KP4", "\033Ot", 0, key_run_command, "w");
    add_keynode("KP5", "\033Ou", 0, key_run_command, "exits");
    add_keynode("KP6", "\033Ov", 0, key_run_command, "e");
    add_keynode("KP8", "\033Ox", 0, key_run_command, "n");
    add_keynode("KP9", "\033Oy", 0, key_run_command, "u");
}
void (*tty_add_walk_binds)() = add_walk_binds;

/*
 * initialize the key binding list
 */
void add_initial_binds()
{
    struct b_init_node {
	char label[8], seq[4];
	void (*funct)();
    };
    static struct b_init_node b_init[] = {
	{ "LF",		"\n",		enter_line },
	{ "Ret",	"\r",		enter_line },
	{ "BS",		"\b",		del_char_left },
	{ "Del",	"\177",		del_char_left },
	{ "Tab",	"\t",		complete_word },
	{ "C-a",	"\001",		beginning_of_line },
	{ "C-b",	"\002",		prev_char },
	{ "C-d",	"\004",		del_char_right },
	{ "C-e",	"\005",		end_of_line },
	{ "C-f",	"\006",		next_char },
	{ "C-k",	"\013",		kill_to_eol },
	{ "C-l",	"\014",		refresh_input_line },
	{ "C-n",	"\016",		next_line },
	{ "C-p",	"\020",		prev_line },
	{ "C-t",	"\024",		transpose_chars },
	{ "C-w",	"\027",		enter_into_history },
	{ "C-z",	"\032",		suspend_powwow },
	{ "M-Tab",	"\033\t",	complete_line },
	{ "M-b",	"\033b",	prev_word },
	{ "M-d",	"\033d",	del_word_right },
	{ "M-f",	"\033f",	next_word },
	{ "M-t",	"\033t",	transpose_words },
	{ "M-u",	"\033u",	upcase_word },
	{ "M-l",	"\033l",	downcase_word },
	{ "M-BS",	"\033\b",	del_word_left },
	{ "M-Del",	"\033\177",	del_word_left },
	{ "",		"",		0 }
    };
    struct b_init_node *p = b_init;
    do {
	add_keynode(p->label, p->seq, 0, p->funct, NULL);
    } while((++p)->seq[0]);
    
    if (cursor_left ) add_keynode("Left" , cursor_left , 0, prev_char, NULL);
    if (cursor_right) add_keynode("Right", cursor_right, 0, next_char, NULL);
    if (cursor_up   ) add_keynode("Up"   , cursor_up   , 0, prev_line, NULL);
    if (cursor_down ) add_keynode("Down" , cursor_down , 0, next_line, NULL);
}
void (*tty_add_initial_binds)() = add_initial_binds;

#ifndef USE_VT100
/*
 * extract termcap 'cap' and strcat it to buf.
 * return the lenght of the extracted string.
 */
static int extract(cap, buf)
char *cap, *buf;
{
    static char *bp;
    char *d = buf + strlen(buf);
    char *s = tgetstr(cap, (bp = d, &bp));
    int len;
    if (!s) return (*bp = 0);
    /*
     * Remove the padding information. We assume that no terminals
     * need padding nowadays. At least it makes things much easier.
     */
    s += strspn(s, "0123456789*");
    for(len = 0; *s; *d++ = *s++, len++)
	if (*s == '$' && *(s + 1) == '<')
	if (!(s = strchr(s, '>')) || !*++s) break;
    *d = 0;
    return len;
}
#endif /* not USE_VT100 */

/*
 * position the cursor using absolute coordinates
 * note: does not flush the output buffer
 */
static void gotoxy(col, line)
int col, line;
{
#ifdef USE_VT100
    tty_printf("\033[%d;%dH", line + 1, col + 1);
#else
    tty_puts(tgoto(curgoto, col, line));
#endif
}
void (*tty_gotoxy)() = gotoxy;

/*
 * optimized cursor movement
 * from (fromcol, fromline) to (tocol, toline)
 * if tocol > 0, (tocol, toline) must lie on editline.
 */
static void gotoxy_opt(fromcol, fromline, tocol, toline)
int fromcol, fromline, tocol, toline;
{
    static char buf[CAPLEN];
    char *cp = buf;
    int cost, i, dist;
    /* First, move vertically to the correct line, then horizontally
     * to the right column. If this turns out to be fewer characters
     * than a direct cursor positioning (tty_gotoxy), use that.
     */
    for (;;) {	/* gotoless */
	if ((i = toline - fromline) < 0) {
	    if (!len_upcur || (cost = -i * len_upcur) >= gotocost)
		break;
	    do {
		strcpy(cp, upcur);
	        cp += len_upcur;
	    } while(++i);
	} else if ((cost = 2 * i)) {	/* lf is mapped to crlf on output */
	    if (cost >= gotocost)
		break;
	    do
	        *cp++ = '\n';
	    while(--i);
	    fromcol = 0;
	}
	if ((i = tocol - fromcol) < 0) {
	    dist = -i * len_leftcur;
	    if (dist <= len_begoln + tocol) {
		if ((cost += dist) > gotocost)
		    break;
		do {
		    strcpy(cp, leftcur);
		    cp += len_leftcur;
		} while(++i);
	    } else {
		if ((cost += len_begoln) > gotocost)
		    break;
		strcpy(cp, begoln);
		cp += len_begoln;
		fromcol = 0;  i = tocol;
	    }
	}
	if (i) {
	    /* if hiliting in effect, just use tty_gotoxy */
	    if (cost + i > gotocost || *edattrbeg)
		break;
	    /* FEATURE: no check is performed whether promptstr consists of
	     *  printable chars only. tty_gotoxy might cost less if it doesn't */
	    if (fromcol < col0 && toline == line0) {
		strcpy(cp, promptstr);
		cp += strlen(cp);
		fromcol = col0;
	    }
	    strncpy(cp, edbuf + (toline - line0) * (cols - 1) + fromcol - col0,
		    tocol - fromcol);
	    cp += tocol - fromcol;
	}
	*cp = 0;
	tty_puts(buf);
	return;
    }
    tty_gotoxy(tocol, toline);
}
void (*tty_gotoxy_opt)() = gotoxy_opt;

/*
 * clear current input line (prompt == 1 if to clear also prompt)
 * cursor is left right after the prompt
 */
static void clear_input_line(prompt)
int prompt;
{
    int newcol = prompt ? 0 : col0;
    if (edlen || prompt) {
	tty_gotoxy_opt(CURCOL(pos), CURLINE(pos), newcol, line0);
	if (line0 < lines - 1)
	    tty_puts(clreoscr);
        else
	    tty_puts(clreoln);
    }
    col0 = newcol;
}
void (*tty_clear_input_line)() = clear_input_line;

/*
 * Redraw the input line and put the cursor at the current position.
 * The cursor is assumed to be directly after the prompt.
 */
static void draw_input_line()
{
    int i, oldline0;
    
    tty_puts(edattrbeg);
    
    if (edlen) {
	oldline0 = line0;
	if (edlen < cols - col0 - 1) {
	    tty_puts(edbuf);
	} else {
	    tty_printf("%.*s", cols - col0 - 1, edbuf);
	    for(i = cols - col0 - 1; i <= edlen; i += cols - 1)
		tty_printf("\n%.*s", cols - 1, edbuf + i);
	}
	line0 = lines - (edlen + col0) / (cols - 1) - 1;
	if (oldline0 < line0) line0 = oldline0;
	tty_gotoxy_opt(CURCOL(edlen), CURLINE(edlen), CURCOL(pos), CURLINE(pos));
    }
}
void (*tty_draw_input_line)() = draw_input_line;

/*
 * delete n characters at current position (the position is unchanged)
 * assert(n <= edlen - pos)
 */
static void delete_nofollow_chars(n)
int n;
{
    int r_cost, d_cost = n, p = pos;
    int nl = p - CURCOL(p);	/* next line's starting pos */
    int cl = CURLINE(p);	/* current line */
    for(d_cost = p + n; d_cost < edlen; d_cost++)
	edbuf[d_cost - n] = edbuf[d_cost];
    for(d_cost -= n; d_cost < edlen; d_cost++)
	edbuf[d_cost] = ' ';
    for( ; ; tty_putc('\n'), p = nl, cl++) {
	d_cost = 0;
	/* FEATURE: ought to be "d_cost = n > gotocost ? -gotocost : -n;"
	 since redraw will need to goto back. Of little importance */
	if ((r_cost = edlen) > (nl += cols - 1))
	    r_cost = nl, d_cost = n + gotocost;
	r_cost -= p;
#ifndef USE_VT100
	/* FEATURE: no clreoln is used (it might cost less in the
					occasion we delete more than one char). Simplicity */
	if (deletecost && deletecost * n + d_cost < r_cost) {
	    for (d_cost = n; d_cost; d_cost--)
	        tty_puts(delchar);
	    if (edlen <= nl)
		break;
	    tty_gotoxy(cols - 1 - n, cl);
	    tty_printf("%.*s", n, edbuf + nl - n - 1);
	} else
#endif /* not USE_VT100 */
	{
	    tty_printf("%.*s", r_cost, edbuf + p);  p += r_cost;
	    if (edlen <= nl) break;
	}
    }
    edbuf[edlen -= n] = 0;
    switch(pos - p) {
     case 1:
	tty_puts(leftcur);
	break;
     case 0:
	break;
     default:
	tty_gotoxy_opt(CURCOL(p), cl, CURCOL(pos), CURLINE(pos));
	break;
    }
}
void (*tty_delete_nofollow_chars)() = delete_nofollow_chars;

/*
 * GH: print a char on current position (overwrite), advance position
 *     from cancan 2.6.3a
 */
static void overtype_follow(c)
char c;
{
    edbuf[pos++] = c;
    tty_putc(c);
    if(!CURCOL(pos))
	tty_putc('\n');
}
void (*tty_overtype_follow)() = overtype_follow;

/*
 * GH: change the position on input line (optmove there, and set pos)
 *     from cancan 2.6.3a
 */
void moveto(new_pos)
int new_pos;
{
    /*
     * FEATURE: the line we are moving to might be less than 0, or greater
     * than lines - 1, if the display is too small to hold the whole editline.
     * In that case, the input line should be (partially) redrawn.
     */
    tty_gotoxy_opt(CURCOL(pos), CURLINE(pos), CURCOL(new_pos), CURLINE(new_pos));
    pos = new_pos;
}
void (*tty_moveto)() = moveto;
  
/*
 * insert n characters at current position. The position is set to after
 * the inserted characters.
 */
static void insert_follow_chars(str, n)
char *str; int n;
{
    int r_cost, i_cost, p = pos;
    int nl = p - CURCOL(p);	/* next line's starting pos */
    int cl = CURLINE(p);	/* current line */
    
    bcopy(edbuf + p, edbuf + p + n, edlen + 1 - p);
    bcopy(str, edbuf + p, n);
    
    edlen += n;  pos += n;
    do {
	i_cost = n;
	if ((r_cost = edlen) > (nl += cols - 1))
	    r_cost = nl, i_cost += insertfinish;
	r_cost -= p;
#ifndef USE_VT100
	/* FEATURE: insert mode is used only when one char is inserted
	 (which is probaly true > 95% of the time). Simplicity */
	if (n == 1 && inscharcost && inscharcost + i_cost < r_cost) {
	    tty_printf("%s%c", inschar, edbuf[p++]);
	    if (edlen > nl) {
		tty_gotoxy(cols - 1, cl);
		tty_puts(clreoln);
	    }
	} else
#endif /* not USE_VT100 */
	{
	    tty_printf("%.*s", r_cost, edbuf + p);  p += r_cost;
	}
	if (edlen < nl)
	    break;
	tty_printf("\n");
	p = nl;
	if (++cl > lines - 1)
	    cl = lines - 1, line0--;
    } while(edlen > nl);
    
    if (p != pos)
	tty_gotoxy_opt(CURCOL(p), cl, CURCOL(pos), CURLINE(pos));
}
void (*tty_insert_follow_chars)() = insert_follow_chars;

/*
 * window size change signal handler:
 * we assume that the bottom line is always the bottom line
 */
static void window_change()
{
    struct winsize wsiz;
    
    /* if ioctl fails or gives silly values, don't change anything */
    if (ioctl(tty_read_fd, TIOCGWINSZ, &wsiz) == 0
	&& wsiz.ws_row && wsiz.ws_col) {
	lines = wsiz.ws_row;
	cols = wsiz.ws_col;
	need_prompt_redraw = 1;
	window_size_changed = 1;
    }
}

void (*tty_window_change)() = window_change;
